/*
 * Copyright 2012-2019 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.samples.petclinic.owner;

import io.opentelemetry.instrumentation.annotations.WithSpan;
import jakarta.persistence.EntityManager;
import jakarta.persistence.PersistenceContext;
import jakarta.validation.Valid;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.samples.petclinic.domain.OwnerValidation;
import org.springframework.stereotype.Controller;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.ui.Model;
import org.springframework.validation.BindingResult;
import org.springframework.web.ErrorResponse;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.servlet.ModelAndView;

import java.net.URI;
import java.util.List;
import java.util.Map;

import static io.opentelemetry.api.GlobalOpenTelemetry.getTracer;

/**
 * @author Juergen Hoeller
 * @author Ken Krebs
 * @author Arjen Poutsma
 * @author Michael Isvy
 */
@Controller
class OwnerController {

	private static final String VIEWS_OWNER_CREATE_OR_UPDATE_FORM = "owners/createOrUpdateOwnerForm";

	private final OwnerRepository owners;

	private OwnerValidation validator;

	@PersistenceContext
	private EntityManager entityManager;

	public OwnerController(OwnerRepository clinicService) {

		this.owners = clinicService;
		var otelTracer = getTracer("OwnerController");
		validator = new OwnerValidation(otelTracer);

	}

	@InitBinder
	public void setAllowedFields(WebDataBinder dataBinder) {
		dataBinder.setDisallowedFields("id");
	}

	@ModelAttribute("owner")
	public Owner findOwner(@PathVariable(name = "ownerId", required = false) Integer ownerId) {
		return ownerId == null ? new Owner() : this.owners.findById(ownerId);
	}

	// @ExceptionHandler(NullPointerException.class)
	// @ResponseStatus(HttpStatus.BAD_REQUEST)
	// public ErrorResponse onIllegaArgumentException(RuntimeException npe){
	// return ErrorResponse.builder(npe,HttpStatus.NOT_FOUND, npe.getMessage())
	// .title("Not found")
	// .type(URI.create("https://api.bookmarks.com/errors/not-found"))
	// .build();
	//
	// }

	@GetMapping("/owners/new")
	public String initCreationForm(Map<String, Object> model) throws NoSuchMethodException {

		// if (model!=null){
		// throw new RuntimeException();
		//
		// }

		Owner owner = new Owner();
		validator.ValidateOwnerWithExternalService(owner);

		model.put("owner", owner);
		return VIEWS_OWNER_CREATE_OR_UPDATE_FORM;
	}

	@GetMapping("/owners/selects")
	public String deathBySelects() throws NoSuchMethodException {

		// if (model!=null){
		// throw new RuntimeException();
		//
		// }
		var owner = this.owners.findById(3);
		for (var pet : owner.getPets()){
			System.out.println(pet.getName());

		}
		return VIEWS_OWNER_CREATE_OR_UPDATE_FORM;

	}
	@GetMapping("/owners/inserts")
	@Transactional
	public String deathByInserts() throws NoSuchMethodException {

		var txn = entityManager.getTransaction();
		txn.begin();
		// if (model!=null){
		// throw new RuntimeException();
		//
		// }
		var owner = this.owners.findById(3);
		for (int i=0;i<3000;i++){
			Pet pet = new Pet();
			PetType type = new PetType();
			type.setName("dog");
			type.setId(1);
			pet.setType(type);
			pet.setName("A"+i);
			owner.addPet(pet);
		}

		this.owners.save(owner);
		txn.commit();
		//model.put("owner", owner);

		return VIEWS_OWNER_CREATE_OR_UPDATE_FORM;
	}

	@PostMapping("/owners/new")
	public String processCreationForm(@Valid Owner owner, BindingResult result) {
		if (result.hasErrors()) {
			return VIEWS_OWNER_CREATE_OR_UPDATE_FORM;
		}
		validator.ValidateOwnerWithExternalService(owner);
		// validator.PerformValidationFlow(owner);

		validator.checkOwnerValidity(owner);
		this.owners.save(owner);
		validator.ValidateUserAccess("admin", "pwd", "fullaccess");
		return "redirect:/owners/" + owner.getId();
	}	

	@GetMapping("/owners/find")
	public String initFindForm() {
		return "owners/findOwners";
	}

	@GetMapping("/owners")
	public String processFindForm(@RequestParam(defaultValue = "1") int page, Owner owner, BindingResult result,
			Model model) {
		validator.ValidateUserAccess("admin", "pwd", "fullaccess");
		// if (owner.getLastName()!=null){
		// throw new RuntimeException();
		//
		// }
		// allow parameterless GET request for /owners to return all records
		if (owner.getLastName() == null) {
			owner.setLastName(""); // empty string signifies broadest possible search
		}

		// find owners by last name
		Page<Owner> ownersResults = findPaginatedForOwnersLastName(page, owner.getLastName());
		if (ownersResults.isEmpty()) {
			// no owners found
			result.rejectValue("lastName", "notFound", "not found");
			return "owners/findOwners";
		}

		if (ownersResults.getTotalElements() == 1) {
			// 1 owner found
			owner = ownersResults.iterator().next();
			return "redirect:/owners/" + owner.getId();
		}

		// multiple owners found
		return addPaginationModel(page, model, ownersResults);
	}

	@WithSpan
	private String test() {
		return "";
	}

	private String addPaginationModel(int page, Model model, Page<Owner> paginated) {
		model.addAttribute("listOwners", paginated);
		List<Owner> listOwners = paginated.getContent();
		model.addAttribute("currentPage", page);
		model.addAttribute("totalPages", paginated.getTotalPages());
		model.addAttribute("totalItems", paginated.getTotalElements());
		model.addAttribute("listOwners", listOwners);
		return "owners/ownersList";
	}

	@WithSpan
	private Page<Owner> findPaginatedForOwnersLastName(int page, String lastname) {
		int pageSize = 25;
		Pageable pageable = PageRequest.of(page - 1, pageSize);
		return owners.findByLastName(lastname, pageable);
	}

	@WithSpan
	private Page<Owner> findPaginatedForOwnersLastNameThree(int page, String lastname) {
		int pageSize = 25;
		Pageable pageable = PageRequest.of(page - 1, pageSize);
		return owners.findByLastName(lastname, pageable);
	}

	private Page<Owner> findPaginatedForOwnersLastNameTwo(int page, String lastname) {
		int pageSize = 25;
		Pageable pageable = PageRequest.of(page - 1, pageSize);
		return owners.findByLastName(lastname, pageable);
	}

	@GetMapping("/owners/{ownerId}/edit")
	public String initUpdateOwnerForm(@PathVariable("ownerId") int ownerId, Model model) {
		Owner owner = this.owners.findById(ownerId);
		model.addAttribute(owner);
		return VIEWS_OWNER_CREATE_OR_UPDATE_FORM;
	}

	@PostMapping("/owners/{ownerId}/edit")
	public String processUpdateOwnerForm(@Valid Owner owner, BindingResult result,
			@PathVariable("ownerId") int ownerId) {
		if (result.hasErrors()) {
			return VIEWS_OWNER_CREATE_OR_UPDATE_FORM;
		}

		owner.setId(ownerId);
		this.owners.save(owner);
		return "redirect:/owners/{ownerId}";
	}

	/**
	 * Custom handler for displaying an owner.
	 * @param ownerId the ID of the owner to display
	 * @return a ModelMap with the model attributes for the view
	 */
	@GetMapping("/owners/{ownerId}")
	public ModelAndView showOwner(@PathVariable("ownerId") int ownerId) {
		ModelAndView mav = new ModelAndView("owners/ownerDetails");
		Owner owner = this.owners.findById(ownerId);
		mav.addObject(owner);

		return mav;
	}

}
